@@ -8,7 +8,7 @@ use std::ffi::OsStr;
 use std::ops::{Deref, DerefMut};
 use std::path::{Path, PathBuf};
 
-use futures::{Future, Stream};
+use futures::{Stream, StreamExt};
 
 use std::task::{Poll, self};
 use std::pin::Pin;
@@ -31,7 +31,6 @@ use crate::sexp::Sexp;
 pub struct Context {
     homedir: Option<PathBuf>,
     sockets: BTreeMap<String, PathBuf>,
-    #[allow(dead_code)] // We keep it around for the cleanup.
     ephemeral: Option<tempfile::TempDir>,
     // XXX: Remove me once hack for Cygwin won't be necessary.
     #[cfg(windows)]
@@ -261,7 +260,7 @@ impl Context {
 
     /// Starts a GnuPG component.
     pub fn start(&self, component: &str) -> Result<()> {
-        self.create_socket_dir()?;
+        let _ = self.create_socket_dir(); // Best effort.
         Self::gpgconf(&self.homedir, &["--launch", component], 1)?;
         Ok(())
     }
@@ -348,17 +347,117 @@ impl Agent {
                           algo: HashAlgorithm, digest: &'a [u8])
         -> Result<crypto::mpi::Signature>
     {
-        SigningRequest::new(&mut self.c, key, algo, digest).await
+        for option in Self::options() {
+            self.send_simple(option).await?;
+        }
+
+        if key.password.is_some() {
+            self.send_simple("OPTION pinentry-mode=loopback").await?;
+        }
+
+        let grip = Keygrip::of(key.public.mpis())?;
+        self.send_simple(format!("SIGKEY {}", grip)).await?;
+        self.send_simple(
+            format!("SETKEYDESC {}",
+                    assuan::escape(&key.password_prompt))).await?;
+
+        let algo = u8::from(algo);
+        let digest = hex::encode(&digest);
+        self.send_simple(format!("SETHASH {} {}", algo, digest)).await?;
+        self.send("PKSIGN")?;
+
+        let mut data = Vec::new();
+        while let Some(r) = self.next().await {
+            match r? {
+                assuan::Response::Ok { .. }
+                | assuan::Response::Comment { .. }
+                | assuan::Response::Status { .. } =>
+                    (), // Ignore.
+                assuan::Response::Inquire { keyword, .. } =>
+                    match (keyword.as_str(), &key.password) {
+                        ("PASSPHRASE", Some(p)) => {
+                            p.map(|p| self.data(p))?;
+                            // Dummy read to send the data.
+                            self.next().await;
+                        },
+                        _ => acknowledge_inquiry(self).await?,
+                    },
+                assuan::Response::Error { ref message, .. } =>
+                    return assuan::operation_failed(self, message).await,
+                assuan::Response::Data { ref partial } =>
+                    data.extend_from_slice(partial),
+            }
+        }
+
+        Sexp::from_bytes(&data)?.to_signature()
     }
 
     /// Decrypts `ciphertext` using `key` with the secret bits managed
     /// by the agent.
     pub async fn decrypt<'a>(&'a mut self,
-                          key: &'a KeyPair,
-                          ciphertext: &'a crypto::mpi::Ciphertext)
+                             key: &'a KeyPair,
+                             ciphertext: &'a crypto::mpi::Ciphertext,
+                             plaintext_len: Option<usize>)
         -> Result<crypto::SessionKey>
     {
-        DecryptionRequest::new(self, key, ciphertext).await
+        for option in Self::options() {
+            self.send_simple(option).await?;
+        }
+
+        if key.password.is_some() {
+            self.send_simple("OPTION pinentry-mode=loopback").await?;
+        }
+
+        let grip = Keygrip::of(key.public.mpis())?;
+        self.send_simple(format!("SETKEY {}", grip)).await?;
+        self.send_simple(format!("SETKEYDESC {}",
+                                  assuan::escape(&key.password_prompt))).await?;
+        self.send("PKDECRYPT")?;
+        let mut padding = true;
+        let mut data = Vec::new();
+        while let Some(r) = self.next().await {
+            match r? {
+                assuan::Response::Ok { .. } |
+                assuan::Response::Comment { .. } =>
+                    (), // Ignore.
+                assuan::Response::Inquire { ref keyword, .. } =>
+                    match (keyword.as_str(), &key.password) {
+                        ("PASSPHRASE", Some(p)) => {
+                            p.map(|p| self.data(p))?;
+                            // Dummy read to send the data.
+                            self.next().await;
+                        },
+                        ("CIPHERTEXT", _) => {
+                            let mut buf = Vec::new();
+                            Sexp::try_from(ciphertext)?.serialize(&mut buf)?;
+                            self.data(&buf)?;
+                            // Dummy read to send the data.
+                            self.next().await;
+                        },
+                        _ => acknowledge_inquiry(self).await?,
+                    },
+                assuan::Response::Status { ref keyword, ref message } =>
+                    if keyword == "PADDING" {
+                        padding = message != "0";
+                    },
+                assuan::Response::Error { ref message, .. } =>
+                    return assuan::operation_failed(self, message).await,
+                assuan::Response::Data { ref partial } =>
+                    data.extend_from_slice(partial),
+            }
+        }
+
+        // Get rid of the safety-0.
+        //
+        // gpg-agent seems to add a trailing 0, supposedly for good
+        // measure.
+        if data.iter().last() == Some(&0) {
+            let l = data.len();
+            data.truncate(l - 1);
+        }
+
+        Sexp::from_bytes(&data)?.finish_decryption(
+            &key.public, ciphertext, plaintext_len, padding)
     }
 
     /// Computes options that we want to communicate.
@@ -404,373 +503,23 @@ impl Agent {
         r.reverse();
         r
     }
-}
 
-struct SigningRequest<'a, 'b, 'c>
-{
-    c: &'a mut assuan::Client,
-    key: &'b KeyPair,
-    algo: HashAlgorithm,
-    digest: &'c [u8],
-    options: Vec<String>,
-    state: SigningRequestState,
-}
-
-impl<'a, 'b, 'c> SigningRequest<'a, 'b, 'c>
-{
-    fn new(c: &'a mut assuan::Client,
-           key: &'b KeyPair,
-           algo: HashAlgorithm,
-           digest: &'c [u8])
-           -> Self {
-        Self {
-            c, key, algo, digest,
-            options: Agent::options(),
-            state: SigningRequestState::Start,
-        }
+    /// Start tracing the data that is sent to the server.
+    ///
+    /// Note: if a tracing function is already registered, this
+    /// replaces it.
+    pub fn trace_data_sent(&mut self, fun: Box<dyn Fn(&[u8]) + Send + Sync>)
+    {
+        self.c.trace_data_sent(fun);
     }
-}
 
-#[derive(Debug)]
-enum SigningRequestState {
-    Start,
-    Options,
-    SigKey,
-    SetKeyDesc,
-    SetHash,
-    PkSign(Vec<u8>),
-}
-
-/// Returns a convenient Err value for use in the state machines
-/// below.
-fn operation_failed<T>(message: &Option<String>) -> Result<T> {
-    Err(Error::OperationFailed(
-        message.as_ref().map(|e| e.to_string())
-            .unwrap_or_else(|| "Unknown reason".into()))
-        .into())
-}
-
-/// Returns a convenient Err value for use in the state machines
-/// below.
-fn protocol_error<T>(response: &assuan::Response) -> Result<T> {
-    Err(Error::ProtocolError(
-        format!("Got unexpected response {:?}", response))
-        .into())
-}
-
-impl<'a, 'b, 'c> Future for SigningRequest<'a, 'b, 'c>
-{
-    type Output = Result<crypto::mpi::Signature>;
-
-    fn poll(self: Pin<&mut Self>, cx: &mut task::Context<'_>) -> Poll<Self::Output> {
-        use self::SigningRequestState::*;
-
-        // The compiler is not smart enough to figure out disjoint borrows
-        // through Pin via DerefMut (which wholly borrows `self`), so unwrap it
-        let Self { c, state, key, options, algo, digest } = Pin::into_inner(self);
-        let mut client = Pin::new(c);
-
-        loop {
-            match state {
-                Start => {
-                    if options.is_empty() {
-                        let grip = Keygrip::of(key.public.mpis())?;
-                        client.send(format!("SIGKEY {}", grip))?;
-                        *state = SigKey;
-                    } else {
-                        let opts = options.pop().unwrap();
-                        client.send(opts)?;
-                        *state = Options;
-                    }
-                },
-
-                Options => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        if let Some(option) = options.pop() {
-                            client.send(option)?;
-                        } else {
-                            let grip = Keygrip::of(key.public.mpis())?;
-                            client.send(format!("SIGKEY {}", grip))?;
-                            *state = SigKey;
-                        }
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                SigKey => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        client.send(
-                            format!("SETKEYDESC {}",
-                                    assuan::escape(&key.password_prompt)))?;
-                        *state = SetKeyDesc;
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                SetKeyDesc => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        let algo = u8::from(*algo);
-                        let digest = hex::encode(&digest);
-                        client.send(format!("SETHASH {} {}", algo, digest))?;
-                        *state = SetHash;
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                SetHash => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        client.send("PKSIGN")?;
-                        *state = PkSign(Vec::new());
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-
-                PkSign(ref mut data) => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        assuan::Response::Data { ref partial } =>
-                            data.extend_from_slice(partial),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        return Poll::Ready(
-                            Sexp::from_bytes(&data)?.to_signature());
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-            }
-        }
-    }
-}
-
-struct DecryptionRequest<'a, 'b, 'c>
-{
-    c: &'a mut assuan::Client,
-    key: &'b KeyPair,
-    ciphertext: &'c crypto::mpi::Ciphertext,
-    options: Vec<String>,
-    state: DecryptionRequestState,
-}
-
-impl<'a, 'b, 'c> DecryptionRequest<'a, 'b, 'c>
-{
-    fn new(c: &'a mut assuan::Client,
-           key: &'b KeyPair,
-           ciphertext: &'c crypto::mpi::Ciphertext)
-           -> Self {
-        Self {
-            c,
-            key,
-            ciphertext,
-            options: Agent::options(),
-            state: DecryptionRequestState::Start,
-        }
-    }
-}
-
-#[derive(Debug)]
-enum DecryptionRequestState {
-    Start,
-    Options,
-    SetKey,
-    SetKeyDesc,
-    PkDecrypt,
-    Inquire(Vec<u8>, bool), // Buffer and padding.
-}
-
-impl<'a, 'b, 'c> Future for DecryptionRequest<'a, 'b, 'c>
-{
-    type Output = Result<crypto::SessionKey>;
-
-    fn poll(mut self: Pin<&mut Self>, cx: &mut task::Context<'_>) -> Poll<Self::Output> {
-        use self::DecryptionRequestState::*;
-
-        // The compiler is not smart enough to figure out disjoint borrows
-        // through Pin via DerefMut (which wholly borrows `self`), so unwrap it
-        let Self { c, state, key, ciphertext, options } = self.deref_mut();
-        let mut client = Pin::new(c);
-
-        loop {
-            match state {
-                Start => {
-                    if options.is_empty() {
-                        let grip = Keygrip::of(key.public.mpis())?;
-                        client.send(format!("SETKEY {}", grip))?;
-                        *state = SetKey;
-                    } else {
-                        let opts = options.pop().unwrap();
-                        client.send(opts)?;
-                        *state = Options;
-                    }
-                },
-
-                Options => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        if let Some(option) = options.pop() {
-                            client.send(option)?;
-                        } else {
-                            let grip = Keygrip::of(key.public.mpis())?;
-                            client.send(format!("SETKEY {}", grip))?;
-                            *state = SetKey;
-                        }
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                SetKey => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        client.send(
-                            format!("SETKEYDESC {}",
-                                    assuan::escape(&key.password_prompt)))?;
-                        *state = SetKeyDesc;
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                SetKeyDesc => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        client.send("PKDECRYPT")?;
-                        *state = PkDecrypt;
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-                PkDecrypt => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Inquire { ref keyword, .. }
-                          if keyword == "CIPHERTEXT" =>
-                            (), // What we expect.
-                        assuan::Response::Comment { .. }
-                        | assuan::Response::Status { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        let mut buf = Vec::new();
-                        Sexp::try_from(*ciphertext)?
-                            .serialize(&mut buf)?;
-                        client.data(&buf)?;
-                        *state = Inquire(Vec::new(), true);
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-
-
-                Inquire(ref mut data, ref mut padding) => match client.as_mut().poll_next(cx)? {
-                    Poll::Ready(Some(r)) => match r {
-                        assuan::Response::Ok { .. }
-                        | assuan::Response::Comment { .. } =>
-                            (), // Ignore.
-                        assuan::Response::Status { ref keyword, ref message } =>
-                            if keyword == "PADDING" {
-                                *padding = message != "0";
-                            },
-                        assuan::Response::Error { ref message, .. } =>
-                            return Poll::Ready(operation_failed(message)),
-                        assuan::Response::Data { ref partial } =>
-                            data.extend_from_slice(partial),
-                        _ =>
-                            return Poll::Ready(protocol_error(&r)),
-                    },
-                    Poll::Ready(None) => {
-                        // Get rid of the safety-0.
-                        //
-                        // gpg-agent seems to add a trailing 0,
-                        // supposedly for good measure.
-                        if data.iter().last() == Some(&0) {
-                            let l = data.len();
-                            data.truncate(l - 1);
-                        }
-
-                        return Poll::Ready(
-                            Sexp::from_bytes(&data)?.finish_decryption(
-                            &key.public, ciphertext, *padding)
-                        );
-                    },
-                    Poll::Pending => return Poll::Pending,
-                },
-            }
-        }
+    /// Start tracing the data that is received from the server.
+    ///
+    /// Note: if a tracing function is already registered, this
+    /// replaces it.
+    pub fn trace_data_received(&mut self, fun: Box<dyn Fn(&[u8]) + Send + Sync>)
+    {
+        self.c.trace_data_received(fun);
     }
 }
 
@@ -782,6 +531,7 @@ impl<'a, 'b, 'c> Future for DecryptionRequest<'a, 'b, 'c>
 pub struct KeyPair {
     public: Key<key::PublicParts, key::UnspecifiedRole>,
     agent_socket: PathBuf,
+    password: Option<crypto::Password>,
     password_prompt: String,
 }
 
@@ -796,6 +546,7 @@ impl KeyPair {
         where R: key::KeyRole
     {
         Ok(KeyPair {
+            password: None,
             password_prompt: format!(
                 "Please enter the passphrase to \
                  unlock the OpenPGP secret key:\n\
@@ -865,6 +616,21 @@ impl KeyPair {
         self.with_password_prompt(prompt)
     }
 
+    /// Supplies a password to unlock the secret key.
+    ///
+    /// This will be used when the secret key operation is performed,
+    /// e.g. when signing or decrypting a message.
+    ///
+    /// Note: This is the equivalent of GnuPG's
+    /// `--pinentry-mode=loopback` and requires explicit opt-in in the
+    /// gpg-agent configuration using the `allow-loopback-pinentry`
+    /// option.  If this is not enabled in the agent, the secret key
+    /// operation will fail.  It is likely only useful during testing.
+    pub fn with_password(mut self, password: crypto::Password) -> Self {
+        self.password = Some(password);
+        self
+    }
+
     /// Changes the password prompt.
     ///
     /// Use this function to give more context to the user when she is
@@ -897,13 +663,38 @@ impl crypto::Signer for KeyPair {
                 | (DSA, PublicKey::DSA { .. })
                 | (EdDSA, PublicKey::EdDSA { .. })
                 | (ECDSA, PublicKey::ECDSA { .. }) => {
-                    let rt = tokio::runtime::Runtime::new()?;
+                    use tokio::runtime::{Handle, Runtime};
 
-                    rt.block_on(async move {
-                        let mut a = Agent::connect_to(&self.agent_socket).await?;
+                    // Connect to the agent and do the signing
+                    // operation.
+                    let do_it = async move {
+                        let mut a =
+                            Agent::connect_to(&self.agent_socket).await?;
                         let sig = a.sign(self, hash_algo, digest).await?;
                         Ok(sig)
-                    })
+                    };
+
+                    // See if the current thread is managed by a tokio
+                    // runtime.
+                    if Handle::try_current().is_err() {
+                        // Doesn't seem to be the case, so it is safe
+                        // to create a new runtime and block.
+                        let rt = Runtime::new()?;
+                        rt.block_on(do_it)
+                    } else {
+                        // It is!  We must not create a second runtime
+                        // on this thread, but instead we will
+                        // delegate this to a new thread.
+                        use crossbeam_utils::thread;
+
+                        thread::scope(|s| {
+                            s.spawn(move |_| {
+                                let rt = Runtime::new()?;
+                                rt.block_on(do_it)
+                            }).join()
+                        }).map_err(map_panic)?
+                            .map_err(map_panic)?
+                    }
                 },
 
             (pk_algo, _) => Err(openpgp::Error::InvalidOperation(format!(
@@ -919,7 +710,7 @@ impl crypto::Decryptor for KeyPair {
     }
 
     fn decrypt(&mut self, ciphertext: &crypto::mpi::Ciphertext,
-               _plaintext_len: Option<usize>)
+               plaintext_len: Option<usize>)
                -> openpgp::Result<crypto::SessionKey>
     {
         use crate::openpgp::crypto::mpi::{PublicKey, Ciphertext};
@@ -928,13 +719,39 @@ impl crypto::Decryptor for KeyPair {
             (PublicKey::RSA { .. }, Ciphertext::RSA { .. })
                 | (PublicKey::ElGamal { .. }, Ciphertext::ElGamal { .. })
                 | (PublicKey::ECDH { .. }, Ciphertext::ECDH { .. }) => {
-                    let rt = tokio::runtime::Runtime::new()?;
+                    use tokio::runtime::{Handle, Runtime};
 
-                    rt.block_on(async move {
-                        let mut a = Agent::connect_to(&self.agent_socket).await?;
-                        let sk = a.decrypt(self, ciphertext).await?;
+                    // Connect to the agent and do the decryption
+                    // operation.
+                    let do_it = async move {
+                        let mut a =
+                            Agent::connect_to(&self.agent_socket).await?;
+                        let sk =
+                            a.decrypt(self, ciphertext, plaintext_len).await?;
                         Ok(sk)
-                    })
+                    };
+
+                    // See if the current thread is managed by a tokio
+                    // runtime.
+                    if Handle::try_current().is_err() {
+                        // Doesn't seem to be the case, so it is safe
+                        // to create a new runtime and block.
+                        let rt = Runtime::new()?;
+                        rt.block_on(do_it)
+                    } else {
+                        // It is!  We must not create a second runtime
+                        // on this thread, but instead we will
+                        // delegate this to a new thread.
+                        use crossbeam_utils::thread;
+
+                        thread::scope(|s| {
+                            s.spawn(move |_| {
+                                let rt = Runtime::new()?;
+                                rt.block_on(do_it)
+                            }).join()
+                        }).map_err(map_panic)?
+                            .map_err(map_panic)?
+                    }
                 },
 
             (public, ciphertext) =>
@@ -946,6 +763,24 @@ impl crypto::Decryptor for KeyPair {
     }
 }
 
+/// Maps a panic of a worker thread to an error.
+///
+/// Unfortunately, there is nothing useful to do with the error, but
+/// returning a generic error is better than panicking.
+fn map_panic(_: Box<dyn std::any::Any + std::marker::Send>) -> anyhow::Error
+{
+    anyhow::anyhow!("worker thread panicked")
+}
+
+/// Helper function to respond to inquiries.
+///
+/// This function doesn't do something useful, but it ends the current
+/// inquiry.
+async fn acknowledge_inquiry(agent: &mut Agent) -> Result<()> {
+    agent.send("END")?;
+    agent.next().await; // Dummy read to send END.
+    Ok(())
+}
 
 #[derive(thiserror::Error, Debug)]
 /// Errors used in this module.
@@ -961,3 +796,91 @@ pub enum Error {
     ProtocolError(String),
 
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use openpgp::{
+        Cert,
+        crypto::{
+            hash::Digest,
+            mpi::Ciphertext,
+            Decryptor,
+            Signer,
+        },
+    };
+
+    /// Asserts that a <KeyPair as Signer> is usable from an async
+    /// context.
+    ///
+    /// Previously, the test died with
+    ///
+    ///     thread 'gnupg::tests::signer_in_async_context' panicked at
+    ///     'Cannot start a runtime from within a runtime. This
+    ///     happens because a function (like `block_on`) attempted to
+    ///     block the current thread while the thread is being used to
+    ///     drive asynchronous tasks.'
+    #[test]
+    fn signer_in_async_context() -> Result<()> {
+        async fn async_context() -> Result<()> {
+            let ctx = match Context::ephemeral() {
+                Ok(c) => c,
+                Err(e) => {
+                    eprintln!("Failed to create ephemeral context: {}", e);
+                    eprintln!("Most likely, GnuPG isn't installed.");
+                    eprintln!("Skipping test.");
+                    return Ok(());
+                },
+            };
+
+            let key = Cert::from_bytes(crate::tests::key("testy-new.pgp"))?
+                .primary_key().key().clone();
+            let mut pair = KeyPair::new(&ctx, &key)?;
+            let algo = HashAlgorithm::default();
+            let digest = algo.context()?.into_digest()?;
+            let _ = pair.sign(algo, &digest);
+            Ok(())
+        }
+
+        let rt = tokio::runtime::Runtime::new()?;
+        rt.block_on(async_context())
+    }
+
+    /// Asserts that a <KeyPair as Decryptor> is usable from an async
+    /// context.
+    ///
+    /// Previously, the test died with
+    ///
+    ///     thread 'gnupg::tests::decryptor_in_async_context' panicked
+    ///     at 'Cannot start a runtime from within a runtime. This
+    ///     happens because a function (like `block_on`) attempted to
+    ///     block the current thread while the thread is being used to
+    ///     drive asynchronous tasks.'
+    #[test]
+    fn decryptor_in_async_context() -> Result<()> {
+        async fn async_context() -> Result<()> {
+            let ctx = match Context::ephemeral() {
+                Ok(c) => c,
+                Err(e) => {
+                    eprintln!("Failed to create ephemeral context: {}", e);
+                    eprintln!("Most likely, GnuPG isn't installed.");
+                    eprintln!("Skipping test.");
+                    return Ok(());
+                },
+            };
+
+            let key = Cert::from_bytes(crate::tests::key("testy-new.pgp"))?
+                .keys().nth(1).unwrap().key().clone();
+            let mut pair = KeyPair::new(&ctx, &key)?;
+            let ciphertext = Ciphertext::ECDH {
+                e: vec![].into(),
+                key: vec![].into_boxed_slice(),
+            };
+            let _ = pair.decrypt(&ciphertext, None);
+            Ok(())
+        }
+
+        let rt = tokio::runtime::Runtime::new()?;
+        rt.block_on(async_context())
+    }
+}
